# ----------------------------------------------------------------------------
# Python Wires
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------


from __future__ import absolute_import

from . import _shell



class _WireWrapper(object):

    def __init__(self, wires_shell):
        self._wires_shell = wires_shell

    def __getattr__(self, name):
        self._wires_shell._wires_instance.wire_context = True
        return getattr(self._wires_shell._wires_instance, name)



class _UnwireWrapper(object):

    def __init__(self, wires_shell):
        self._wires_shell = wires_shell

    def __getattr__(self, name):
        self._wires_shell._wires_instance.wire_context = False
        return getattr(self._wires_shell._wires_instance, name)



class _WiresSingleton(object):

    def __init__(self):
        wires_shell = _shell.WiresShell()
        self.wire = _WireWrapper(wires_shell)
        self.unwire = _UnwireWrapper(wires_shell)


_WIRE_SINGLETON = _WiresSingleton()

wire = _WIRE_SINGLETON.wire
unwire = _WIRE_SINGLETON.unwire


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Python Call Wires.
"""

from __future__ import absolute_import

from . import _instance



class WiresShell(object):

    def __init__(self):

        self._wires_instance = _instance.WiresInstance()


    @property
    def wire(self):

        self._wires_instance.wire_context = True
        return self._wires_instance


    @property
    def unwire(self):

        self._wires_instance.wire_context = False
        return self._wires_instance


    def __getattr__(self, name):

        return getattr(self._wires_instance, name)


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Simple callable.
"""


from __future__ import absolute_import

import logging
import sys
from traceback import print_exception



class Callable(object):

    """
    Callable with a minimal API.
    Summary:
    - Has zero or more handlers: functions/callables added to it.
    - Fired by calling it, like a function.
    - Firing it calls all associated handlers, passing them any
      arguments given to call.
    """

    def __init__(self, name, wires, logger_name='wires'):

        self._name = name
        self._wires = wires
        self._logger_name = logger_name

        # See `_log_handler_failure` below.
        self.use_log = True

        # Handler functions/callables.
        self._functions = []


    def calls_to(self, function):

        """
        Adds/removes `function` as a handler.
        """

        if not callable(function):
            raise ValueError('argument not callable: %r' % (function,))

        # TODO: explain
        if self._wires.wire_context:
            self._functions.append(function)
        else:
            try:
                self._functions.remove(function)
            except ValueError as e:
                e.args = ('unknown function %r: %s' % (function, e),)
                raise


    def __call__(self, *args, **kwargs):

        # Calls all handler functions.

        for function in self._functions:
            try:
                function(*args, **kwargs)
            except Exception as e:
                # Catching exceptions here is critical to ensure decoupling
                # callables from callees.
                self._log_handler_failure(function, e)


    def _log_handler_failure(self, function, e):

        # Try to produce a useful message including:
        # - The callable name.
        # - The function name.
        # - The raised execption.

        # `self.use_log` is used to prevent usage of the logging system:
        # necessary if any logging handler depends on calling us, which may
        # lead to us failing again, the handler failing again, ad infinitum
        #
        # Possible values:
        # - True: logging system will be used.
        # - False: outputs to sys.stderr.
        # - None: No output will be produced.
        #         (useful if logging system captures stderr)

        handler_name = self._handler_name(function)
        msg = '%r calling %r failed: %r' % (self._name, handler_name, e)
        if self.use_log:
            logger = logging.getLogger(self._logger_name)
            logger.error(msg)
            logger.exception(e)
        elif self.use_log is not None:
            sys.stderr.write(msg+'\n')
            print_exception(type(e), e, e.__traceback__, file=sys.stderr)


    @staticmethod
    def _handler_name(function):

        # Return the best possible name for the function.
        # Will be formatted like "module_name.function_name".

        function_name = getattr(function, '__qualname__', None)
        if function_name is None:
            # Older Python versions to not support __qualname__.
            function_name = getattr(function, '__name__', 'unnamed-callable')

        module_name = getattr(function, '__module__', 'unnamed-module')

        return '%s.%s' % (module_name, function_name)


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Python Call Wires.
"""

from __future__ import absolute_import

from . import _callable



class WiresInstance(object):

    """
    Callable based event manager with a minimal API.
    Summary:
    - Events are attributes of Wires, created dynamically, as needed.
    - Events have zero or more handlers: functions/callables added to them.
    - Events are fired by calling them, like functions.
    - Firing an event calls all associated handlers, passing them the same
      arguments the "fire event" call was given.

    Usage example:

    >>> em = Wires()

    # Tell the event manager to call our lambda when 'my_event' is fired.
    >>> em.my_event.calls(lambda: print('event handler #1'))

    # Adding another callable for 'my_event':
    >>> em.my_event.calls(lambda: print('event handler #2'))

    # Trigger 'my_event'
    >>> em.my_event()
    event handler #1
    event handler #2
    """

    def __init__(self):

        # Tracks known Callable instances:
        # - Keys are callable names (my dynamic attributes).
        # - Values are Callable objects.

        self._callables = {}

        # TODO: explain this
        self.wire_context = True


    def __getattr__(self, name):

        # Called on attribute access, returns an event object.
        # Either uses an tracked one or creates new one, tracking it.

        try:
            return self._callables[name]
        except KeyError:
            new_callable = _callable.Callable(name, self)
            self._callables[name] = new_callable
            return new_callable


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------


from __future__ import absolute_import

import unittest

from wires import wire, unwire

from . import helpers



class TestWiresAPI(unittest.TestCase):

    """
    Minimal API tests for the Wires singleton.
    """

    @staticmethod
    def test_unwired_call_does_not_fail():
        """
        Calling an unwired call works. Does nothing, but works.
        """
        wire.unwired_call()


    def test_wiring_non_callable_raises_value_error(self):
        """
        Wiring a call to a non-callable raises ValueError.
        The exception argument (message):
        - Starts with "argument not callable: ".
        - Contains repr(argument).
        """
        for non_callable in (None, True, 42, 2.3, (), [], {}, set()):

            with self.assertRaises(ValueError) as cm:
                wire.this.calls_to(non_callable)

            exception_args = cm.exception.args
            self.assertEqual(len(exception_args), 1)

            msg = exception_args[0]
            self.assertTrue(
                msg.startswith('argument not callable: '),
                'wrong exception message: %r' % (msg,),
            )
            self.assertIn(
                repr(non_callable),
                msg,
                'missing argument repr in message: %r' % (msg,),
            )


    @staticmethod
    def test_wiring_callable_works():
        """
        Wiring a callable works.
        """
        wire.this.calls_to(lambda: None)


    @staticmethod
    def test_wiring_unwiring_works():
        """
        Wiring and then unwiring to the same callable works.
        """
        test_callable = lambda: None

        wire.this.calls_to(test_callable)
        unwire.this.calls_to(test_callable)


    def test_unwiring_unknown_callable_raises_value_error(self):
        """
        Unwiring an unknown callable raises a ValueError.
        The exception argument (message):
        - Starts with "unknown function ".
        """
        test_callable = lambda: None

        with self.assertRaises(ValueError) as cm:
            unwire.this.calls_to(test_callable)

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)

        msg = exception_args[0]
        self.assertTrue(
            msg.startswith('unknown function '),
            'wrong exception message: %r' % (msg,),
        )



class TestWiresUtilization(helpers.CallTrackerAssertMixin, unittest.TestCase):

    """
    Utilization tests for the Wires singleton.
    """

    def test_wire_call(self):
        """
        Wires the `this` call and calls it.
        Checks the wired callable was called once with the correct arguments.
        """
        tracker = helpers.CallTracker()

        wire.this.calls_to(tracker)
        wire.this()

        self.assertEqual(tracker.call_count, 1, 'call count mismatch')
        self.assertEqual(tracker.call_args, [
            ((), {},),
        ], 'call argument mismatch')


    def test_double_wire_call(self):
        """
        Wires the `this` call twice and calls it.
        Checks the wired callable was called twice with the correct arguments.
        """
        tracker = helpers.CallTracker()

        wire.this.calls_to(tracker)
        wire.this.calls_to(tracker)
        wire.this()

        self.assertEqual(tracker.call_count, 2, 'call count mismatch')
        self.assertEqual(tracker.call_args, [
            ((), {},),
            ((), {},),
        ], 'call argument mismatch')


    def test_multi_wired_call(self):
        """
        Wires the `this` call to multiple callables and calls it.
        Checks that each of the wired callables was called once with the correct
        arguments.
        """
        num_wirings = 10

        trackers = [helpers.CallTracker() for _ in range(num_wirings)]

        for tracker in trackers:
            wire.this.calls_to(tracker)

        wire.this()

        for tracker in trackers:
            self.assertEqual(tracker.call_count, 1, 'call count mismatch')
            self.assertEqual(tracker.call_args, [
                ((), {},),
            ], 'call argument mismatch')


    def test_wire_call_unwire_call(self):
        """
        Wires the `this` call to a callable and calls it.
        Checks the wired callable was called once with the correct arguments.
        Unwires the `this` callable and calls it again.
        Checks the wired callable wasn't called again.
        """
        tracker = helpers.CallTracker()

        wire.this.calls_to(tracker)
        wire.this()

        self.assert_single_call_no_args(tracker)

        unwire.this.calls_to(tracker)
        wire.this()

        self.assert_single_call_no_args(tracker)


# ----------------------------------------------------------------------------
